"""Markdown documentation generation for SpinASM instructions."""

# ruff: noqa: E501

from __future__ import annotations

from spinasm_lsp.docs.markdown import Arg, ArgList, Instruction

INSTRUCTIONS: dict[str, Instruction] = {
    "ABSA": Instruction(
        name="ABSA",
        args=ArgList([]),
        description="""**`ABSA`** loads the accumulator with the absolute value of the accumulator.""",
        parameter_description=None,
        operation="|ACC| ­> ACC",
        coding="00000000000000000000000000001001",
        example="""    absa            ; Absolute value of ACC -> ACC
""",
    ),
    "AND": Instruction(
        name="AND",
        args=ArgList(
            [
                Arg("MASK", 24, ["Binary", "Hex ($000000-$FFFFFF)", "Symbolic"]),
            ]
        ),
        description="""**`AND MASK`** will perform a bit wise "and" of the current `ACC` and the 24-­bit `MASK` specified within the instruction word. The instruction might be used to load a constant into `ACC` provided `ACC` contains `$FFFFFF` or to clear `ACC` if `MASK` equals `$000000`. (see also the pseudo opcode section)""",
        parameter_description=None,
        operation="ACC & MASK",
        coding="MMMMMMMMMMMMMMMMMMMMMMMM000001110",
        example="""AMASK EQU $F0FFFF

;----------------------------------------
    or ­­­­­­­­­­­­­­­­­­$FFFFFF                        ; Set all bits within ACC
    and $FFFFFE                       ; Clear LSB
    and %01111111_11111111_11111111   ; Clear MSB
    and AMASK                         ; Clear ACC[19..16]
    and $0                            ; Clear ACC
""",
    ),
    "CHO RDA": Instruction(
        name="CHO RDA",
        args=ArgList(
            [
                Arg("N", 2, ["LFO select: SIN0,SIN1,RMP0,RMP1"]),
                Arg("C", 6, ["Binary", "Bit flags"]),
                Arg("D", 16, ["Real (S.15)", "Symbolic"]),
            ]
        ),
        description="""**`CHO RDA, N, C, D`**, like the `RDA` instruction, will read a sample from the delay ram, multiply it by a coefficient and add the product to the previous content of `ACC`. However, in contrast to `RDA` the coefficient is not explicitly embedded within the instruction and the effective delay ram address is not solely determined by the address parameter. Instead, both values are modulated by the selected LFO at run time, for an in depth explanation please consult the FV­1 datasheet alongside with application note AN­0001. `CHO RDA` is a very flexible and powerful instruction, especially useful for delay line modulation effects such as chorus or pitch shifting.

The coefficient field of the `CHO` instructions are used as control bits to select various aspects of the LFO. These bits can be set using predefined flags that are `OR`ed together to create the required bit field. For a sine wave LFO (`SIN0` or `SIN1`), valid flags are:

`SIN COS REG COMPC COMPA`

While for a ramp LFO (RMP0 and RMP1), valid flags are:

`REG COMPC COMPA RPTR2 NA`

These flags are defined as:

| Flag  | HEX value | Description                                              |
| ----- | --------- | -------------------------------------------------------- |
| SIN   | $0        | Select SIN output (default) (Sine LFO only)              |
| COS   | $1        | Select COS output (Sine LFO only)                        |
| REG   | $2        | Save the output of the LFO into an internal LFO register |
| COMPC | $4        | Complement the coefficient (1-coeff)                     |
| COMPA | $8        | Complement the address offset from the LFO               |
| RPTR2 | $10       | Select the ramp+1/2 pointer (Ramp LFO only)              |
| NA    | $20       | Select x-fade coefficient and do not add address offset  |""",
        parameter_description=None,
        operation="See description",
        coding="00CCCCCC0NNAAAAAAAAAAAAAAAA10100",
        example="""; A chorus
Delay MEM 4097                          ; Chorus delay line
Amp EQU 8195                            ; Amplitude for a 4097 sample delay line
Freq EQU 51                             ; Apx. 2Hz at 32kHz sampling rate

; Setup SIN LFO 0
    skp run,cont                        ; Skip if not first iteration
    wldr 0,Freq,Amp                     ; Setup SIN LFO 0

cont:
    sof 0,0                             ; Clear ACC
    rdax ADCL,1.0                       ; Read left ADC * 1.0
    wra Delay,0                         ; Write to delay line, clear ACC
    cho rda,RMP0,COMPC|REG,Delay        ; See application note AN-0001
    cho rda,RMP0,,Delay+1               ; for detailed examples and explanation
    wra Temp,0                          ;
    cho rda,RMP0,COMPC|RPTR2,Delay      ;
    cho rda,RMP0,RPTR2,Delay+1          ;
    cho sof,RMP0,NA|COMPC,0             ;
    cho rda,RMP0,NA,Temp                ;
    wrax DACL,0                         ; Result to DACL and clear ACC
""",
    ),
    "CHO RDAL": Instruction(
        name="CHO RDAL",
        args=ArgList(
            [
                Arg("N", 2, ["LFO select: SIN0,COS0,SIN1,COS1,RMP0,RMP1"]),
            ]
        ),
        description="""**`CHO RDAL, N`** will read the current value of the selected LFO into `ACC`.""",
        parameter_description=None,
        operation="LFO * 1 ­> ACC",
        coding="110000100NN000000000000000010100",
        example="""cho rdal,SIN0           ; Read LFO S0 into ACC
wrax DACL,0             ; Result to DACL and clear ACC
""",
    ),
    "CHO SOF": Instruction(
        name="CHO SOF",
        args=ArgList(
            [
                Arg("N", 2, ["LFO select: SIN0,SIN1,RMP0,RMP1"]),
                Arg("C", 6, ["Binary", "Bit flags"]),
                Arg("D", 16, ["Real (S.15)", "Symbolic"]),
            ]
        ),
        description="""**`CHO SOF, N, C, D`**, like the `SOF` instruction, will multiply ACC by a coefficient and add the constant `D` to the result. However, in contrast to `SOF` the coefficient is not explicitly embedded within the instruction. Instead, based on the selected LFO and the 6 bit vector `C`, the coefficient is picked from a list of possible coefficients available within the LFO block of the FV­1. For an in depth explanation please consult the FV­-1 datasheet alongside with application note AN­0001. `CHO SOF` is a very flexible and powerful instruction, especially useful for the cross fading portion of pitch shift algorithms.

Please see `CHO RDA` for a description of field flags.""",
        parameter_description=None,
        operation="See description",
        coding="10CCCCCC0NNDDDDDDDDDDDDDDDD10100",
        example="""; Pitch shift
Delay MEM 4096                          ; Pitch shift delay line
Temp MEM 1                              ; Temporary storage
Amp EQU 4096                            ; RAMP LFO amplitude (4096 samples)
Freq EQU -8192                          ; RAMP LFO frequency

; Setup RAMP LFO 0
    skp run,cont                        ; Skip if not first iteration
    wldr 0,Freq,Amp                     ; Setup RAMP LFO 0

cont:
    sof 0,0                             ; Clear ACC
    rdax ADCL,1.0                       ; Read left ADC * 1.0
    wra Delay,0                         ; Write to delay line, clear ACC
    cho rda,RMP0,COMPC|REG,Delay        ; See application note AN-0001
    cho rda,RMP0,,Delay+1               ; for detailed examples and explanation
    wra Temp,0                          ;
    cho rda,RMP0,COMPC|RPTR2,Delay      ;
    cho rda,RMP0,RPTR2,Delay+1          ;
    cho sof,RMP0,NA|COMPC,0             ;
    cho rda,RMP0,NA,Temp                ;
    wrax DACL,0                         ; Result to DACL and clear ACC
""",
    ),
    "CLR": Instruction(
        name="CLR",
        args=ArgList([]),
        description="""**`CLR`** will clear the accumulator.""",
        parameter_description=None,
        operation="0 ­> ACC",
        coding="00000000000000000000000000001110",
        example="""    clr                 ; Clear ACC
    rdax ADCL,1.0       ; Read left ADC
                        ;-----------------
    ....                ; ...Left channel processing...
                        ;-----------------
    wrax DACL,0         ; Result to DACL and clear ACC
""",
    ),
    "EXP": Instruction(
        name="EXP",
        args=ArgList(
            [
                Arg("C", 16, ["Real (S1.14)", "Hex ($0000-$FFFF)", "Symbolic"]),
                Arg("D", 11, ["Real (S.10)", "Symbolic"]),
            ]
        ),
        description="""**`EXP C, D`** will multiply `2^ACC` with `C` and add the constant `D` to the result.

Since `ACC` (in it’s role as the destination for the `EXP` instruction) is limited to linear values from 0 to
+0.99999988, the `EXP` instruction is limited to logarithmic `ACC` values (in it’s role as the source operand
for the `EXP` instruction) from –16 to 0. Like the LOG instruction, `EXP` will treat the `ACC` content as a
S4.19 number. Positive logarithmic `ACC` values will be clipped to +0.99999988 which is the most positive
linear value that can be represented within the accumulator.

`D` is intended to allow the linear `ACC` to be offset by a constant in the range from –1 to +0.9990234375""",
        parameter_description=None,
        operation="C * EXP(ACC) + D",
        coding="CCCCCCCCCCCCCCCCDDDDDDDDDDD01100",
        example="""exp 0.8,0
""",
    ),
    "JAM": Instruction(
        name="JAM",
        args=ArgList(
            [
                Arg("N", 1, ["RAMP LFO select: (0, 1)"]),
            ]
        ),
        description="""**`JAM N`** will reset the selected RAMP LFO to its starting point.""",
        parameter_description=None,
        operation="0 ­> RAMP LFO N",
        coding="0000000000000000000000001N010011",
        example="""jam 0           ; Force ramp 0 LFO to it's starting position
""",
    ),
    "LDAX": Instruction(
        name="LDAX",
        args=ArgList(
            [
                Arg("ADDR", 6, ["Decimal (0-63)", "Hex ($0-$3F)", "Symbolic"]),
            ]
        ),
        description="""**`LDAX ADDR`** loads `ACC` with the contents of the addressed register `ADDR`.""",
        parameter_description=None,
        operation="REG[ADDR]­> ACC",
        coding="00000000000000000000000000000101",
        example="""    ldax adcl       ; ADC left input -> ACC
""",
    ),
    "LOG": Instruction(
        name="LOG",
        args=ArgList(
            [
                Arg("C", 16, ["Real (S1.14)", "Hex ($0000-$FFFF)", "Symbolic"]),
                Arg("D", 11, ["Real (S4.6)", "Symbolic"]),
            ]
        ),
        description="""**`LOG C, D`** will multiply the Base2 `LOG` of the current absolute value in `ACC` with `C` and add the constant `D` to the result.

It is important to note that the `LOG` function returns a fixed point number in `S4.19` format instead of the standard `S.23` format, which in turn means that the most negative Base2 `LOG` value is -16.

The `LOG` instruction can handle absolute linear accumulator values from 0.99999988 to 0.00001526 which translates to a dynamic range of apx. 96dB.

`D` is an offset to be added to the logarithmic value in the range of -16 to + 15.999998.""",
        parameter_description=None,
        operation="C * LOG(|ACC|) + D",
        coding="CCCCCCCCCCCCCCCCDDDDDDDDDDD01011",
        example="""log 1.0,0
""",
    ),
    "MAXX": Instruction(
        name="MAXX",
        args=ArgList(
            [
                Arg("ADDR", 6, ["Decimal (0-63)", "Hex ($0-$3F)", "Symbolic"]),
                Arg("C", 16, ["Real (S1.14)", "Hex ($8000-$0000-$7FFF)", "Symbolic"]),
            ]
        ),
        description="""**`MAXX ADDR, C`** will compare the absolute value of `ACC` versus C times the absolute value of the register pointed to by `ADDR`. If the absolute value of `ACC` is larger `ACC` will be loaded with `|ACC|`, otherwise the accumulator becomes overwritten by `|REG[ADDR] * C|`.""",
        parameter_description="""In order to simplify the MAXX syntax, see the list of predefined symbols for all registers within the FV-1 register file.""",
        operation="MAX( |REG[ADDR] * C| , |ACC| )",
        coding="CCCCCCCCCCCCCCCC00000AAAAAA01001",
        example="""; Peak follower
;
Peak EQU 32          ; Peak hold register

    sof 0,0          ; Clear ACC 
    rdax ADCL,1.0    ; Read left ADC
    maxx Peak,1.0    ; Keep larger absolute value in ACC

; For a peak meter insert decay code here...

    wrax Peak,0.0    ; Save (new) peak and clear ACC
""",
    ),
    "MULX": Instruction(
        name="MULX",
        args=ArgList(
            [
                Arg("ADDR", 6, ["Decimal (0-63)", "Hex ($0-$3F)", "Symbolic"]),
            ]
        ),
        description="""**`MULX ADDR`** will multiply `ACC` by the value of the register pointed to by `ADDR`. An important application of the `MULX` instruction is squaring the content of `ACC`, which combined with a single order LP is especially useful in calculating the RMS value of an arbitrary waveform.""",
        parameter_description="""In order to simplify the `MULX` syntax, see the list of predefined symbols for all registers within the FV-1 register file.""",
        operation="ACC * REG[ADDR]",
        coding="000000000000000000000AAAAAA01010",
        example="""; RMS conversion
Tmp_LP EQU 32        ; Temporary register for first order LP

    sof 0,0          ; Clear ACC 
    rdax ADCL,1.0    ; Read left ADC
                     ; RMS calculation = ACC^2 -> first order LP

    mulx ADCL        ; ACC^2
    rdfx Tmp_LP,x.x  ; First order...
    wrax Tmp_LP,1.0  ; ...LP filter

; At this point ACC holds the RMS value of the input
""",
    ),
    "NOT": Instruction(
        name="NOT",
        args=ArgList([]),
        description="""**`NOT`** will negate all bit positions within accumulator thus performing a 1’s complement.""",
        parameter_description=None,
        operation="/ACC ­> ACC",
        coding="11111111111111111111111100010000",
        example="""    not             ; 1's comp ACC
""",
    ),
    "OR": Instruction(
        name="OR",
        args=ArgList(
            [
                Arg("MASK", 24, ["Binary", "Hex ($000000-$FFFFFF)", "Symbolic"]),
            ]
        ),
        description="""**`OR MASK`** will perform a bit wise "or" of the current `ACC` and the 24-­bit `MASK` specified within the instruction word. The instruction might be used to load a constant into `ACC` provided `ACC` contains `$000000`.""",
        parameter_description=None,
        operation="ACC | MASK",
        coding="MMMMMMMMMMMMMMMMMMMMMMMM000001111",
        example="""0MASK EQU $0F0000

;----------------------------------------
    sof 0,0                           ; Clear all bits within ACC
    or $1                             ; Set LSB
    or %10000000_00000000_00000000    ; Set MSB
    or 0MASK                          ; Set ACC[19..16]
    and %S=[15..8]                    ; Set ACC[15..8]
""",
    ),
    "RDA": Instruction(
        name="RDA",
        args=ArgList(
            [
                Arg(
                    "ADDR",
                    (1) + 15,
                    ["Decimal (0-32767)", "Hex ($0-$7FFF)", "Symbolic"],
                ),
                Arg("C", 11, ["Real (S1.9)", "Hex ($400-$000-$3FF)", "Symbolic"]),
            ]
        ),
        description="""**`RDA ADDR, C`** will fetch the sample [ADDR] from the delay RAM, multiply it by `C`, and add the result to the previous content of `ACC`. This multiply-accumulate operation is probably the most popular operation found in DSP algorithms.""",
        parameter_description=None,
        operation="SRAM[ADDR] * C + ACC",
        coding="CCCCCCCCCCCAAAAAAAAAAAAAAAA00000",
        example="""Delay MEM 1024 
Coeff EQU 1.55
Tmp   EQU $2000

    rda 1000,1.9
    rda Delay+20,Coeff
    rda Tmp,-2
    rda $7FFF,$7FF
""",
    ),
    "RDAX": Instruction(
        name="RDAX",
        args=ArgList(
            [
                Arg("ADDR", 6, ["Decimal (0-63)", "Hex ($0-$3F)", "Symbolic"]),
                Arg("C", 16, ["Real (S1.14)", "Hex ($8000-$0000-$7FFF)", "Symbolic"]),
            ]
        ),
        description="""**`RDAX ADDR, C`** will fetch the value contained in `[ADDR]` from the register file, multiply it with `C` and add the result to the previous content of `ACC`. This multiply accumulate is probably the most popular operation found in DSP algorithms.""",
        parameter_description="""In order to simplify the `RDAX` syntax, see the list of predefined symbols for all registers within the FV-1 register file.""",
        operation="C * REG[ADDR] + ACC",
        coding="CCCCCCCCCCCCCCCC00000AAAAAA00100",
        example="""; Crude mono 
;
    sof 0,0          ; Clear ACC 
    rdax ADCL,0.5    ; Get ADCL value and divide it by two 
    rdax ADCR,0.5    ; Get ADCR value, divide it by two 
                     ; and add to the half of ADCL 
    wrax DACL,1.0    ; Result to DACL 
    wrax DACR,0      ; Result to DACR and clear ACC
""",
    ),
    "RDFX": Instruction(
        name="RDFX",
        args=ArgList(
            [
                Arg("ADDR", 6, ["Decimal (0-63)", "Hex ($0-$3F)", "Symbolic"]),
                Arg("C", 16, ["Real (S1.14)", "Hex ($8000-$0000-$7FFF)", "Symbolic"]),
            ]
        ),
        description="""**`RDFX ADDR, C`** will subtract the value of the register pointed to by `ADDR` from `ACC`, multiply the result by `C` and then add the value of the register pointed to by `ADDR`. `RDFX` is an extremely powerful instruction in that it represents the major portion of a single order low pass filter.""",
        parameter_description="""In order to simplify the `RDFX` syntax, see the list of predefined symbols for all registers within the FV-1 register file.""",
        operation="(ACC­REG[ADDR])*C + REG[ADDR]",
        coding="CCCCCCCCCCCCCCCC00000AAAAAA00101",
        example="""; Single order LP filter
Tmp_LP EQU 32        ; Temporary register for first order LP

    ldax ADCL        ; Read left ADC
    rdfx Tmp_LP,x.x  ; First order...
    wrax Tmp_LP,1.0  ; ...LP filter
    wrax DACL,0      ; Result to DACL and clear ACC
""",
    ),
    "RMPA": Instruction(
        name="RMPA",
        args=ArgList(
            [
                Arg("C", 11, ["Real (S1.9)", "Hex ($400-$000-$3FF)", "Symbolic"]),
            ]
        ),
        description="""**`RMPA C`** provides indirect delay line addressing in that the delay line address of the sample to be multiplied by `C` is not explicitly given in the instruction itself but contained within the pointer register `ADDR_PTR` (absolute address 24 within the internal register file.)

`RMPA` will fetch the indirectly addressed sample from the delay ram, multiply it by `C` and add the result to the previous content of `ACC`.""",
        parameter_description=None,
        operation="SRAM[PNTR[N]] * C + ACC",
        coding="CCCCCCCCCCC000000000001100000001",
        example="""; Crude variable delay line addressing
    sof 0,0             ; Clear ACC
    rdax POT1,1.0       ; Read POT1 value
    wrax ADDR_PTR,0     ; Write value to pointer register, clear ACC
    rmpa 1.0            ; Read sample from delay line
    wrax DACL,0         ; Result to DACL and clear ACC
""",
    ),
    "SKP": Instruction(
        name="SKP",
        args=ArgList(
            [
                Arg("CMASK", 5, ["Binary", "Hex ($00-$1F)", "Symbolic"]),
                Arg("N", 6, ["Decimal (1-63)", "Label"]),
            ]
        ),
        description="""**`SKP CMASK, N`** allows conditional program execution. The FV-1 features five condition flags that can be used to conditionally skip the next `N` instructions. The selection of which condition flag(s) must be asserted in order to skip the next `N` instructions is made by the five-bit condition mask `CMASK`. Only if all condition flags that correspond to a logic "1" within `CMASK` are asserted are the following `N` instructions skipped. The individual bits within `CMASK` correspond to the FV-1 condition flags as follows:

| CMASK | Flag | Description                                                                                                                                                                                                                                                                                             |
| ----- | ---- | ------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------- |
| b4    | RUN  | The RUN flag is cleared after the program has executed for the first time after it was loaded into the internal program memory. The purpose of the RUN flag is to allow the program to initialize registers and LFOs during the first sample iteration then to skip those initializations from then on. |
| b3    | ZRC  | The ZRC flag is asserted if the sign of ACC and PACC is different, a condition that indicates a Zero Crossing.                                                                                                                                                                                          |
| b2    | ZRO  | Z is asserted if ACC = 0                                                                                                                                                                                                                                                                                |
| b1    | GEZ  | GEZ is asserted if ACC >= 0                                                                                                                                                                                                                                                                             |
| b0    | NEG  | N is asserted if ACC is negative                                                                                                                                                                                                                                                                        |""",
        parameter_description="""Maybe the most efficient way to define the condition mask is using its symbolic representation. In order to simplify the `SKP` syntax, SPINAsm has a predefined set of symbols which correspond to the name of the individual condition flags. (RUN, ZRC, ZRO, GEZ, NEG). Although most of the condition flags are mutually exclusive, SPINAsm allows you to specify more than one condition flag to become evaluated simply by separating multiple predefined symbols by the "|" character. Accordingly, `skp ZRC|N, 6` would skip the following six instructions in case of a zero crossing to a negative value.""",
        operation="CMASK N",
        coding="CCCCCNNNNNN000000000000000010001",
        example="""; A bridge rectifier
;
    sof 0,0          ; Clear ACC 
    rdax ADCL,1.0    ; Read from left ADC channel
    skp GEZ,pos      ; Skip next instruction if ACC >= 0
    sof -1.0,0       ; Make ACC positive
pos:
    wrax DACL,0      ; Result to DACL, clear ACC
    rdax ADCL,1.0    ; Read from left ADC channel
    skp N,neg        ; Skip next instruction if ACC < 0
    sof -1.0,0       ; Make ACC negative
neg:
    wrax 0,DACR      ; Result to DACR, clear
""",
    ),
    "SOF": Instruction(
        name="SOF",
        args=ArgList(
            [
                Arg("C", 16, ["Real (S1.14)", "Hex ($0000-$FFFF)", "Symbolic"]),
                Arg("D", 11, ["Real (S.10)", "Symbolic"]),
            ]
        ),
        description="""**`SOF C, D`** will multiply the current value in `ACC` with `C` and will then add the constant `D` to the result.

Please note the absence of an integer entry format for `D`. This is not by mistake but it should emphasize that `D` is not intended to become used for integer arithmetic. The reason for this instruction is that the 11 bit constant `D` would be placed into `ACC` left justified or in other words 13 bits shifted to the left. `D` is intended to offset `ACC` by a constant in the range from -1 to +0.9990234375.""",
        parameter_description=None,
        operation="C * ACC + D",
        coding="CCCCCCCCCCCCCCCCDDDDDDDDDDD01101",
        example="""Off EQU 1.0

; Halve way rectifier­­­
    sof 0,0                 ; Clear ACC
    rdax ADCL,1.0           ; Read from left ADC channel
    sof 1.0,­-Off            ; Subtract offset
    sof 1.0,Off             ; Add offset
""",
    ),
    "WLDR": Instruction(
        name="WLDR",
        args=ArgList(
            [
                Arg("N", 1, ["RAMP LFO select: (0, 1)"]),
                Arg(
                    "F",
                    16,
                    [
                        "Decimal (-16384-32768)",
                        "Hex ($4000-$000-$7FFF)",
                        "Symbolic",
                    ],
                ),
                Arg("A", 2, ["Decimal (512, 1024, 2048, 4096)", "Symbolic"]),
            ]
        ),
        description="""**`WLDR N, F, A`** will load frequency and amplitude control values into the selected RAMP LFO. (0 or 1) This instruction is intended to setup the selected RAMP LFO which is typically done within the first sample iteration after a new program became loaded. As a result `WLDR` will in most cases be used in combination with a `SKP RUN` instruction. For a more detailed description regarding the frequency and amplitude control values see application note AN­0001.""",
        parameter_description=None,
        operation="See description",
        coding="01NFFFFFFFFFFFFFFFF000000AA10010",
        example="""Amp EQU 4096            ; LFO will module a 4096 sample delay line
Freq EQU $100           ; 
;------------------------

; Setup RAMP LFO 0       ;
    skp run,start       ; Skip next instruction if not first iteration
    wldr 0,Freq,Amp     ; Setup RAMP LFO 0

start: and 0,0          ;
    ....                ;
    ....                ;
""",
    ),
    "WLDS": Instruction(
        name="WLDS",
        args=ArgList(
            [
                Arg("N", 1, ["SIN LFO select: (0, 1)"]),
                Arg("F", 9, ["Decimal (0-511)", "Hex ($000-$1FF)", "Symbolic"]),
                Arg("A", 15, ["Decimal (0-32767)", "Hex ($0000-$7FFF)", "Symbolic"]),
            ]
        ),
        description="""**`WLDS N, F, A`** will load frequency and amplitude control values into the selected SIN LFO (0 or 1). This instruction is intended to setup the selected SIN LFO which is typically done within the first sample iteration after a new program is loaded. As a result `WLDS` will in most cases be used in combination with a `SKP RUN` instruction. For a more detailed description regarding the frequency and amplitude control values see application note AN­0001.""",
        parameter_description=None,
        operation="See description",
        coding="00NFFFFFFFFFAAAAAAAAAAAAAAA10010",
        example="""Amp EQU 8194            ; Amplitude for a 4097 sample delay line
Freq EQU 51             ; Apx. 2Hz at 32kHz sampling rate
;------------------------

; Setup SIN LFO 0       ;
    skp run,start       ; Skip next instruction if not first iteration
    wlds 0,Freq,Amp     ; Setup SIN LFO 0

start: sof 0,0          ;
    ....                ;
    ....                ;
""",
    ),
    "WRA": Instruction(
        name="WRA",
        args=ArgList(
            [
                Arg(
                    "ADDR",
                    (1) + 15,
                    ["Decimal (0-32767)", "Hex ($0-$7FFF)", "Symbolic"],
                ),
                Arg("C", 11, ["Real (S1.9)", "Hex ($400-$000-$3FF)", "Symbolic"]),
            ]
        ),
        description="""**`WRA ADDR, C`** will store `ACC` to the delay ram location addressed by `ADDR` and then multiply `ACC` by `C`.""",
        parameter_description=None,
        operation="ACC­>SRAM[ADDR], ACC * C",
        coding="CCCCCCCCCCCAAAAAAAAAAAAAAAA00010",
        example="""Delay MEM 1024 
Coeff EQU 0.5

    sof 0,0             ; Clear ACC
    rdax ADCL,1.0       ; Read left ADC
    wra Delay,Coeff     ; Write to start of delay line, halve ACC
    rda Delay#,Coeff    ; Add half of the sample from the end of the delay line
    wrax DACL,0         ; Result to DACL and clear ACC
""",
    ),
    "WRAP": Instruction(
        name="WRAP",
        args=ArgList(
            [
                Arg(
                    "ADDR",
                    (1) + 15,
                    ["Decimal (0-32767)", "Hex ($0-$7FFF)", "Symbolic"],
                ),
                Arg("C", 11, ["Real (S1.9)", "Hex ($400-$000-$3FF)", "Symbolic"]),
            ]
        ),
        description="""**`WRAP ADDR, C`** will store `ACC` to the delay ram location addressed by `ADDR` then multiply `ACC` by `C` and finally add the content of the `LR` register to the product. Please note that the `LR` register contains the last sample value read from the delay ram memory. This instruction is typically used for all­pass filters in a reverb program.""",
        parameter_description=None,
        operation="ACC­>SRAM[ADDR], (ACC*C) + LR",
        coding="CCCCCCCCCCCAAAAAAAAAAAAAAAA00011",
        example="""    rda ap1#,kap        ; Read output of all-pass 1 and multiply it by kap
    wrap ap1,-kap       ; Write ACC to input of all-pass 1 and do
                        ; ACC*(-kap)+ap1# (ap1# is in LR register)
""",
    ),
    "WRAX": Instruction(
        name="WRAX",
        args=ArgList(
            [
                Arg("ADDR", 6, ["Decimal (0-63)", "Hex ($0-$3F)", "Symbolic"]),
                Arg("C", 16, ["Real (S1.14)", "Hex ($8000-$0000-$7FFF)", "Symbolic"]),
            ]
        ),
        description="""**`WRAX ADDR, C`** will save the current value in `ACC` to `[ADDR]` and then multiply `ACC` by `C`. This instruction can be used to write `ACC` to one DAC channel while clearing `ACC` for processing the next audio channel.""",
        parameter_description="""In order to simplify the `WRAX` syntax, see the list of predefined symbols for all registers within the FV­1.""",
        operation="ACC­>REG[ADDR], C * ACC",
        coding="CCCCCCCCCCCCCCCC00000AAAAAA00110",
        example="""; Stereo processing
;
    rdax ADCL,1.0    ; Read left ADC into previously cleared ACC
                     ;---------------
    ....             ; ...left channel processing
                     ;---------------
    wrax DACL,0      ; Result to DACL and clear ACC for right channel processing
    rdax ADCR,1.0    ; Read right ADC into previously cleared ACC
                     ;---------------
    ....             ; ...right channel processing
                     ;---------------
    wrax DACR,0      ; Result to DACR and clear ACC for left channel processing
""",
    ),
    "WRHX": Instruction(
        name="WRHX",
        args=ArgList(
            [
                Arg("ADDR", 6, ["Decimal (0-63)", "Hex ($0-$3F)", "Symbolic"]),
                Arg("C", 16, ["Real (S1.14)", "Hex ($8000-$0000-$7FFF)", "Symbolic"]),
            ]
        ),
        description="""**`WRHX ADDR, C`** stores the current `ACC` value in the register pointed to by `ADDR`, then multiplies `ACC` by `C`. Finally, the previous content of `ACC` (`PACC`) is added to the product. `WRHX` is an extremely powerful instruction; when combined with `RDFX`, it forms a single order high pass shelving filter.""",
        parameter_description="""In order to simplify the `WRHX` syntax, see the list of predefined symbols for all registers within the FV-1 register file.""",
        operation="ACC -> REG[ADDR], (ACC * C) + PACC",
        coding="CCCCCCCCCCCCCCCC00000AAAAAA00111",
        example="""; Single order HP shelving filter
Tmp_HP EQU 32       ; Temporary register for first order HP

;----------------------------------------

    sof 0,0         ; Clear ACC
    rdax ADCL,1.0   ; Read left ADC
    rdfx Tmp_HP,x.x ; First order HP...
    wrhx Tmp_HP,y.y ; ...shelving filter
    wrax DACL,0     ; Result to DACL and clear ACC
""",
    ),
    "WRLX": Instruction(
        name="WRLX",
        args=ArgList(
            [
                Arg("ADDR", 6, ["Decimal (0-63)", "Hex ($0-$3F)", "Symbolic"]),
                Arg("C", 16, ["Real (S1.14)", "Hex ($8000-$0000-$7FFF)", "Symbolic"]),
            ]
        ),
        description="""**`WRLX ADDR, C`** stores the current `ACC` value into the register pointed to by `ADDR`, then subtracts `ACC` from the previous content of `ACC` (`PACC`). The difference is then multiplied by C and finally `PACC` is added to the result. `WRLX` is an extremely powerful instruction in that when combined with `RDFX`, it forms a single order low-pass shelving filter.""",
        parameter_description="""In order to simplify the `WRLX` syntax, see the list of predefined symbols for all registers within the FV-1 register file.""",
        operation="ACC -> REG[ADDR], (PACC-ACC) * C + PACC",
        coding="CCCCCCCCCCCCCCCC00000AAAAAA01000",
        example="""; Single order LP shelving filter
Tmp_LP EQU 32       ; Temporary register for first order LP

;----------------------------------------

    sof 0,0         ; Clear ACC
    rdax ADCL,1.0   ; Read left ADC
    rdfx Tmp_LP,x.x ; First order LP...
    wrlx Tmp_LP,y.y ; ...shelving filter
    wrax DACL,1.0   ; Result to DACL and clear ACC
""",
    ),
    "XOR": Instruction(
        name="XOR",
        args=ArgList(
            [
                Arg("MASK", 24, ["Binary", "Hex ($000000-$FFFFFF)", "Symbolic"]),
            ]
        ),
        description="""**`XOR MASK`** will perform a bit wise "xor" of the current `ACC` and the 24­-bit `MASK` specified within the instruction word. The instruction will invert `ACC` provided `MASK` equals `$FFFFFF`. (see also the pseudo opcode section).""",
        parameter_description=None,
        operation="ACC ^ MASK",
        coding="MMMMMMMMMMMMMMMMMMMMMMMM000010000",
        example="""XMASK EQU $AAAAAA

;----------------------------------------
    sof 0,0                           ; Clear all bits within ACC
    xor $0                            ; Set all ACC bits
    xor %01010101_01010101_01010101   ; Invert all even numbered bits
    xor XMASK                          ; Invert all odd numbered bits
""",
    ),
}



"""Tools for generating Markdown documentation."""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import cached_property
from typing import List, Literal


class MarkdownGenerator(ABC):
    """An abstract class for dataclasses that can be converted into markdown strings."""

    @property
    @abstractmethod
    def markdown(self) -> str:
        """A markdown documentation string."""

    def __str__(self) -> str:
        return self.markdown


@dataclass
class Arg(MarkdownGenerator):
    """Metadata for an argument to a SPINAsm instruction."""

    name: str
    width: int
    formats: list[str]

    @property
    def markdown(self) -> str:
        return f"{self.name}: {' | '.join(self.formats)}"


class ArgList(MarkdownGenerator, List[Arg]):
    """A collection of arguments for a SPINAsm instruction."""

    @property
    def markdown(self) -> str:
        return ", ".join([arg.markdown for arg in self])


@dataclass
class Instruction(MarkdownGenerator):
    """Generatable Markdown documentation for a SPINAsm instruction."""

    name: str
    args: ArgList
    description: str
    operation: str
    coding: str
    example: str
    parameter_description: str | None = None

    @cached_property
    def markdown(self) -> str:
        """A markdown documentation string."""
        md = MarkdownString()

        md.add_paragraph(self.description.strip())

        md.add_heading("Operation", level=4)
        md.add_paragraph(f"`{self.operation}`")

        md.add_heading("Parameters", level=4)
        if not self.args:
            md.add_paragraph("None.")
        else:
            md.add_table(
                cols=["Name", "Width", "Entry formats, range"],
                rows=[
                    [arg.name, f"{arg.width} Bit", "<br>".join(arg.formats)]
                    for arg in self.args
                ],
            )
        if self.parameter_description:
            md.add_paragraph(self.parameter_description)

        md.add_heading("Example", level=4)
        md.add_codeblock(self.example.strip(), language="assembly")

        md.add_horizontal_rule()
        md.add_paragraph(
            "*Adapted from Spin Semiconductor SPINAsm & FV-1 Instruction Set reference "
            "manual. Copyright 2008 by Spin Semiconductor.*"
        )

        return str(md)


@dataclass
class Assembler(MarkdownGenerator):
    """Generatable Markdown documentation for a SPINAsm assembler."""

    name: str
    description: str
    example: str
    example_remarks: str = ""

    @cached_property
    def markdown(self) -> str:
        """A markdown documentation string."""
        md = MarkdownString()

        md.add_paragraph(self.description.strip())

        md.add_heading("Example", level=4)
        md.add_codeblock(self.example.strip(), language="assembly")

        if self.example_remarks:
            md.add_paragraph(self.example_remarks.strip())

        md.add_horizontal_rule()
        md.add_paragraph(
            "*Adapted from Spin Semiconductor SPINAsm & FV-1 Instruction Set reference "
            "manual. Copyright 2008 by Spin Semiconductor.*"
        )

        return str(md)


class MarkdownTable(MarkdownGenerator):
    def __init__(
        self,
        cols: list[str],
        rows: list[list[str]],
        justify: list[Literal["left", "center", "right"]] | None = None,
    ):
        ncol = len(cols)
        if any([len(row) != ncol for row in rows]):
            raise ValueError(f"All row lengths must match col length of {ncol}.")
        if justify is not None and len(justify) != ncol:
            raise ValueError("There must be one justify position per column.")

        if justify is None:
            self.justify = ["left"] * ncol
        self.cols = cols
        self.rows = rows

    @property
    def markdown(self) -> str:
        header = " | ".join(self.cols)

        separators = []
        for just in self.justify:
            if just == "left":
                separators.append(":-")
            elif just == "right":
                separators.append("-:")
            else:
                separators.append(":-:")
        separator = " | ".join(separators)
        rows = "\n".join([" | ".join(row) for row in self.rows])

        return f"{header}\n{separator}\n{rows}"


class MarkdownString:
    def __init__(self):
        self._content = ""

    def __str__(self):
        return self._content

    def _add_line(self, s: str):
        self._content += f"\n{s}\n"

    def add_heading(self, title: str, level: int):
        if level < 1 or level > 4:
            raise ValueError("Level must be > 0 and < 5.")
        self._add_line(f"{'#' * level} {title}")

    def add_horizontal_rule(self):
        self._add_line("-" * 24)

    def add_paragraph(self, s: str):
        self._add_line(s)

    def add_table(self, cols: list[str], rows: list[list[str]]):
        self._add_line(str(MarkdownTable(cols, rows)))

    def add_codeblock(self, source: str, language: str | None = None):
        block = f"```{language}\n{source}\n```"
        self._add_line(block)



"""Markdown documentation generation for SpinASM assemblers."""

# ruff: noqa: E501

from __future__ import annotations

from spinasm_lsp.docs.markdown import Assembler

ASSEMBLERS: dict[str, Assembler] = {
    "EQU": Assembler(
        name="EQU",
        description="""**`EQU`** allows one to define symbolic operands in order to increase the readability of the source code. Technically an `EQU` statement such as:

```assembly
Name EQU Value [;Comment]
```

will cause SPINAsm to replace any occurrence of the literal "Name" by the literal "Value" within each instruction line during the assembly process excluding the comment portion of an instruction line.

With the exception of blanks, any printable character is allowed within the literal "Name". However there are restrictions: "Name" must be an unique string, is limited to 32 characters and the first character must be a letter excluding the "+" and "­" signs and the "!" character.

The reason for not allowing these characters being the first character of "Name" is that any symbolic operand may be prefixed with a sign or the "!" negation operator within the instruction line. The assembler will then perform the required conversion of the operand while processing the individual instruction lines.

There is another, not syntax related, restriction when using symbolic operands defined by an `EQU` statement: Predefined symbols. As given in the end of the manual there is a set of predefined symbolic operands which should be omitted as "Name" literals within an `EQU` statement. It is not that these predefined symbols are prohibited, it is just that using them within an `EQU` statement will overwrite their predefined value.

With the literal "Value" things are slightly more complicated since its format has to comply with the syntactical rules defined for the operand type it is to represent. Although it is suggested to place `EQU` statements at the beginning of the source code file, this is not mandatory. However, the `EQU` statement has to be defined before the literal "Name" can be used as a symbolical operand within an instruction line.

#### Remark
SPINAsm has no way of performing range checking while processing the EQU statement. This is because the operand type of value is not known to SPINAsm at the time the EQU statement is processed. As a result, range checking is performed when assembling the instruction line in which "Name" is to be replaced by "Value".
""",
        example="""
Attn      EQU 0.5      ; 0.5 = -6dB attenuation
Tmp_Reg   EQU 63       ; Temporary register within register file
Tmp_Del   EQU $2000    ; Temporary memory location within delay ram
;
;------------------------------
sof       0,0          ; Clear ACC
rda       Tmp_Del,Attn ; Load sample from delay ram $2000,
                       ; multiply it by 0.5 and add ACC content
wrax      Tmp_Reg,1.0  ; Save result to Tmp_Reg but keep it in ACC
wrax      DACL,0       ; Move ACC to DAC left (predefined symbol)
                       ; and then clear ACC
""",
        example_remarks="""If `Tmp_Del` was accidentally replaced by `Tmp_Reg` within the `rda` instruction line, SPINAsm would not detect this semantic error – simply because using `Tmp_Reg` would be syntactically correct.""",
    ),
    "MEM": Assembler(
        name="MEM",
        description="""**`MEM`** allows the user to partition the delay ram memory into individual blocks. A memory block declared by the statement

```assembly
Name `MEM` Size [;Comment]
```

can be referenced by `Name` from within an instruction line. `Name` has to comply with the same syntactical rules previously defined with the EQU statement, "Size" is an unsigned integer in the range of 1 to 32768 which might be entered either in decimal or in hexadecimal.

Besides the explicit identifier `Name` the assembler defines two additional implicit identifiers, `Name#` and `Name^`. `Name` refers to the first memory location within the memory block, whereas `Name#` refers to the last memory location. The identifier `Name^` references the middle of the memory block, or in other words its center. If a memory block of size 1 is defined, all three identifiers will address the same memory location. In case the memory block is of size 2, `Name` and `Name^` will address the same memory location, if the size is an even number the memory block cannot exactly be halved – the midpoint `Name^` will be calculated as: `size MOD 2`.

Optionally all three identifiers can be offset by a positive or negative integer which is entered in decimal. Although range checking is performed when using offsets, there is no error generated if the result of the address calculation exceeds the address range of the memory block. This is also true for those cases in which the result will "wrap around" the physical 32k boundary of the delay memory. However, a warning will be issued in order to alert the user regarding the out of range condition.

Mapping the memory blocks to their physical delay ram addresses is solely handled by SPINAsm. The user has no possibility to explicitly force SPINAsm to place a certain memory block to a specific physical address range. This of course does not mean that the user has no control over the layout of the delay ram at all: Knowing that SPINAsm will map memory blocks in the order they become defined within the source file, the user can implicitly control the memory map of the delay ram.
""",
        example="""DelR      MEM  1024    ; Right channel delay line
DelL      MEM  1024    ; Left channel delay line
                       ;
;------------------------------
sof       0,0          ; Clear ACC
rdax      ADCL,1.0     ; Read in left ADC
wra       DelL,0.25    ; Save it to the start of the left delay
                       ; line and keep a -12dB replica in ACC
rdax      DelL^+20,0.25; Add sample from "center of the left delay
                       ; line + 20 samples" times 0.25 to ACC
rdax      DelL#,0.25   ; Add sample from "end of the left delay
                       ; line" times 0.25 to ACC
rdax      DelL-512,0.25; Add sample from "start of the left delay
                       ; line - 512 samples" times 0.25 to ACC
""",
        example_remarks="""#### Remark
At this point the result of the address calculation will reference a sample from outside the `DelL` memory block. While being syntactically correct, the instruction might not result in what the user intended. In order to make the user aware of that potential semantic error, a warning will be issued.

```assembly
wrax      DACL,0       ; Result to DACL, clear ACC
                       ;
rdax      ADCR,1.0     ; Read in right ADC
wra       DelR,0.25    ; Save it to the start of the right delay
                       ; line and keep a -12dB replica in ACC
rdax      DelR^-20,0.25; Add sample from center of the right delay
                       ; line - 20 samples times 0.25 to ACC
rdax      DelR#,0.25   ; Add sample from end of the right delay line
                       ; times 0.25 to ACC
rdax      DelR-512,0.25; Add sample from start of the right delay
                       ; line - 512 samples times 0.25 to ACC
```

#### Remark
At this point the result of the address calculation will reference a sample from outside the `DelR` memory block. And even worse than the previous case: This time the sample be fetched from delay ram address 32256 which will contain a sample that is apx. 1 second old!

Again, syntactically correct but most likely a semantic error – warnings will be issued.

```assembly
wrax DACR,0         ; Result to DACR, clear ACC
```
""",
    ),
}



"""The SPINAsm language parser."""

from __future__ import annotations

import bisect
import copy
from typing import Literal, TypedDict

import lsprotocol.types as lsp
from asfv1 import fv1parse


class Symbol(TypedDict):
    """
    The token specification used by asfv1.

    Note that we exclude EOF tokens, as they are ignored by the LSP.
    """

    type: Literal[
        "ASSEMBLER",
        "INTEGER",
        "LABEL",
        "TARGET",
        "MNEMONIC",
        "OPERATOR",
        "FLOAT",
        "ARGSEP",
    ]
    txt: str
    stxt: str
    val: int | float | None


class Token:
    """
    A parsed token.

    Parameters
    ----------
    symbol : Symbol
        The symbol parsed by asfv1 representing the token.
    start : lsp.Position
        The start position of the token in the source file.
    end : lsp.Position, optional
        The end position of the token in the source file. If not provided, the end
        position is calculated based on the width of the symbol's stxt.

    Attributes
    ----------
    symbol : Symbol
        The symbol parsed by asfv1 representing the token.
    range : lsp.Range
        The location range of the token in the source file.
    """

    def __init__(
        self, symbol: Symbol, start: lsp.Position, end: lsp.Position | None = None
    ):
        if end is None:
            width = len(symbol["stxt"])
            end = lsp.Position(line=start.line, character=start.character + width)

        self.symbol: Symbol = symbol
        self.range: lsp.Range = lsp.Range(start=start, end=end)

    def __repr__(self) -> str:
        return self.symbol["stxt"]

    def concatenate(self, other: Token) -> Token:
        """
        Concatenate by merging with another token, in place.

        In practice, this is used for the multi-word opcodes that are parsed as separate
        tokens: CHO RDA, CHO RDAL, and CHO SOF.
        """
        if any(
            symbol_type not in ("MNEMONIC", "LABEL")
            for symbol_type in (self.symbol["type"], other.symbol["type"])
        ):
            raise TypeError("Only MNEMONIC and LABEL symbols can be concatenated.")

        self.symbol["txt"] += f" {other.symbol['txt']}"
        self.symbol["stxt"] += f" {other.symbol['stxt']}"
        self.range.end = other.range.end
        return self

    def _clone(self) -> Token:
        """Return a clone of the token to avoid mutating the original."""
        return copy.deepcopy(self)

    def without_address_modifier(self) -> Token:
        """
        Create a clone of the token with the address modifier removed.
        """
        if not str(self).endswith("#") and not str(self).endswith("^"):
            return self

        token = self._clone()
        token.symbol["stxt"] = token.symbol["stxt"][:-1]
        token.range.end.character -= 1

        return token


class TokenRegistry:
    """A registry of tokens and their positions in a source file."""

    def __init__(self, tokens: list[Token] | None = None) -> None:
        self._prev_token: Token | None = None

        """A dictionary mapping program lines to all Tokens on that line."""
        self._tokens_by_line: dict[int, list[Token]] = {}

        """A dictionary mapping token names to all matching Tokens in the program."""
        self._tokens_by_name: dict[str, list[Token]] = {}

        for token in tokens or []:
            self.register_token(token)

    def register_token(self, token: Token) -> None:
        """Add a token to the registry."""
        # Handle multi-word CHO instructions by merging the second token with the first
        # and skipping the second token.
        if str(self._prev_token) == "CHO" and str(token) in ("RDA", "RDAL", "SOF"):
            self._prev_token.concatenate(token)  # type: ignore
            return

        if token.range.start.line not in self._tokens_by_line:
            self._tokens_by_line[token.range.start.line] = []

        # Store the token on its line
        self._tokens_by_line[token.range.start.line].append(token)
        self._prev_token = token

        # Store user-defined tokens together by name. Other token types could be stored,
        # but currently there's no use case for retrieving their positions.
        if token.symbol["type"] in ("LABEL", "TARGET"):
            # Tokens are stored by name without address modifiers, so that e.g. Delay#
            # and Delay can be retrieved with the same query. This allows for renaming
            # all instances of a memory token.
            token = token.without_address_modifier()

            if str(token) not in self._tokens_by_name:
                self._tokens_by_name[str(token)] = []

            self._tokens_by_name[str(token)].append(token)

    def get_matching_tokens(self, token_name: str) -> list[Token]:
        """Retrieve all tokens with a given name in the program."""
        return self._tokens_by_name.get(token_name.upper(), [])

    def get_token_at_position(self, position: lsp.Position) -> Token | None:
        """Retrieve the token at the given position."""
        if position.line not in self._tokens_by_line:
            return None

        line_tokens = self._tokens_by_line[position.line]
        token_starts = [t.range.start.character for t in line_tokens]
        token_ends = [t.range.end.character for t in line_tokens]

        idx = bisect.bisect_left(token_starts, position.character)

        # The index returned by bisect_left points to the start value >= character. This
        # will either be the first character of the token or the start of the next
        # token. First check if we're out of bounds, then shift left unless we're at the
        # first character of the token.
        if idx == len(line_tokens) or token_starts[idx] != position.character:
            idx -= 1

        # If the col falls after the end of the token, we're not inside a token.
        if position.character > token_ends[idx]:
            return None

        return line_tokens[idx]


class SPINAsmParser(fv1parse):
    """A modified version of fv1parse optimized for use with LSP."""

    sym: Symbol | None

    def __init__(self, source: str):
        self.diagnostics: list[lsp.Diagnostic] = []
        """A list of diagnostic messages generated during parsing."""

        self.definitions: dict[str, lsp.Range] = {}
        """A dictionary mapping symbol names to their definition location."""

        self.current_character: int = 0
        """The current column in the source file."""

        self.previous_character: int = 0
        """The last visitied column in the source file."""

        self.token_registry = TokenRegistry()
        """A registry of tokens and their positions in the source file."""

        super().__init__(
            source=source,
            clamp=True,
            spinreals=False,
            # Ignore the callbacks in favor of overriding their callers
            wfunc=lambda *args, **kwargs: None,
            efunc=lambda *args, **kwargs: None,
        )

        # Track which symbols were defined at initialization, e.g. registers and LFOs
        self.constants: list[str] = list(self.symtbl.keys())
        # Keep an unchanged copy of the original source
        self._source: list[str] = self.source.copy()

    def __mkopcodes__(self):
        """
        No-op.

        Generating opcodes isn't needed for LSP functionality, so we'll skip it.
        """

    def _record_diagnostic(
        self, msg: str, line: int, character: int, severity: lsp.DiagnosticSeverity
    ):
        """Record a diagnostic message for the LSP."""
        self.diagnostics.append(
            lsp.Diagnostic(
                range=lsp.Range(
                    start=lsp.Position(line, character=character),
                    end=lsp.Position(line, character=character),
                ),
                message=msg,
                severity=severity,
                source="SPINAsm",
            )
        )

    def parseerror(self, msg: str, line: int | None = None):
        """Override to record parsing errors as LSP diagnostics."""
        if line is None:
            line = self.prevline

        # Offset the line from the parser's 1-indexed line to the 0-indexed line
        self._record_diagnostic(
            msg,
            line=line - 1,
            character=self.current_character,
            severity=lsp.DiagnosticSeverity.Error,
        )

    def scanerror(self, msg: str):
        """Override to record scanning errors as LSP diagnostics."""
        self._record_diagnostic(
            msg,
            line=self.current_line,
            character=self.current_character,
            severity=lsp.DiagnosticSeverity.Error,
        )

    def parsewarn(self, msg: str, line: int | None = None):
        """Override to record parsing warnings as LSP diagnostics."""
        if line is None:
            line = self.prevline

        # Offset the line from the parser's 1-indexed line to the 0-indexed line
        self._record_diagnostic(
            msg,
            line=line - 1,
            character=self.current_character,
            severity=lsp.DiagnosticSeverity.Warning,
        )

    @property
    def sline(self):
        return self._sline

    @sline.setter
    def sline(self, value):
        """Update the current line and reset the column."""
        self._sline = value

        # Reset the column to 0 when we move to a new line
        self.previous_character = self.current_character
        self.current_character = 0

    @property
    def current_line(self):
        """Get the zero-indexed current line."""
        return self.sline - 1

    @property
    def previous_line(self):
        """Get the zero-indexed previous line."""
        return self.prevline - 1

    def __next__(self):
        """Parse the next symbol and update the column and definitions."""
        super().__next__()
        if self.sym["type"] == "EOF":
            return

        self._update_column()

        token = Token(
            symbol=self.sym,
            start=lsp.Position(
                line=self.current_line, character=self.current_character
            ),
        )
        self.token_registry.register_token(token)

        base_token = token.without_address_modifier()
        is_user_definable = base_token.symbol["type"] in ("LABEL", "TARGET")
        is_defined = str(base_token) in self.jmptbl or str(base_token) in self.symtbl

        if (
            is_user_definable
            and not is_defined
            # Labels appear before their target definition, so override when the target
            # is defined.
            or base_token.symbol["type"] == "TARGET"
        ):
            self.definitions[str(base_token)] = base_token.range

    def _update_column(self):
        """Set the current column based on the last parsed symbol."""
        current_line_txt = self._source[self.current_line]
        current_symbol = self.sym.get("txt", None) or ""

        self.previous_character = self.current_character
        try:
            # Start at the current column to skip previous duplicates of the symbol
            self.current_character = current_line_txt.index(
                current_symbol, self.current_character
            )
        except ValueError:
            self.current_character = 0

    def parse(self) -> SPINAsmParser:
        """Parse and return the parser."""
        super().parse()
        return self



from __future__ import annotations

from spinasm_lsp.docs.assemblers import ASSEMBLERS
from spinasm_lsp.docs.instructions import INSTRUCTIONS
from spinasm_lsp.docs.markdown import Instruction, MarkdownGenerator

# Opcodes where the first argument is considered part of the instruction rather than
# an argument, which requires some special handling.
MULTI_WORD_INSTRUCTIONS = ("CHO RDA", "CHO RDAL", "CHO SOF")


class DocumentationManager:
    """A manager for case-insensitive documentation lookups."""

    instructions = INSTRUCTIONS
    assemblers = ASSEMBLERS
    data: dict[str, MarkdownGenerator] = {**INSTRUCTIONS, **ASSEMBLERS}

    def __getitem__(self, key: str) -> str:
        return str(self.data[key.upper()])

    def get_markdown(self, key: str, default: str = "") -> str:
        return str(self.data.get(key.upper(), default))

    def get_instruction(self, key: str) -> Instruction | None:
        return self.instructions.get(key.upper(), None)

    def __contains__(self, key: str) -> bool:
        return self.data.__contains__(key.upper())

    def __iter__(self):
        return iter(self.data)


__all__ = ["DocumentationManager"]



"""The SPINAsm Language Server Protocol implementation."""

from __future__ import annotations

from functools import lru_cache
from typing import Any

from lsprotocol import types as lsp
from pygls.server import LanguageServer

from spinasm_lsp import __version__
from spinasm_lsp.docs import MULTI_WORD_INSTRUCTIONS, DocumentationManager
from spinasm_lsp.parser import SPINAsmParser


@lru_cache(maxsize=1)
def _parse_document(source: str) -> SPINAsmParser:
    """
    Parse a document and return the parser.

    Parser are cached based on the source code to speed up subsequent parsing.
    """
    return SPINAsmParser(source).parse()


class SPINAsmLanguageServer(LanguageServer):
    def __init__(self, *args, **kwargs) -> None:
        self._prev_parser: SPINAsmParser | None = None
        self.documentation = DocumentationManager()

        super().__init__(*args, name="spinasm-lsp", version=__version__, **kwargs)

    def debug(self, msg: Any) -> None:
        """Log a debug message."""
        # MessageType.Debug is a proposed feature of 3.18.0, and isn't fully supported
        # yet.
        self.show_message_log(str(msg), lsp.MessageType.Log)

    def info(self, msg: Any) -> None:
        """Log an info message."""
        self.show_message_log(str(msg), lsp.MessageType.Info)

    def warning(self, msg: Any) -> None:
        """Log a warning message."""
        self.show_message_log(str(msg), lsp.MessageType.Warning)

    def error(self, msg: Any) -> None:
        """Log an error message."""
        self.show_message_log(str(msg), lsp.MessageType.Error)

    async def get_parser(self, uri: str) -> SPINAsmParser:
        """Return a parser for the document, caching if possible."""
        document = self.workspace.get_text_document(uri)
        parser = _parse_document(document.source)

        # Skip publishing diagnostics if the parser is unchanged
        if parser is not self._prev_parser:
            self.publish_diagnostics(document.uri, parser.diagnostics)
            self._prev_parser = parser

        return parser


server = SPINAsmLanguageServer(max_workers=5)


@server.feature(lsp.TEXT_DOCUMENT_DID_CHANGE)
async def did_change(
    ls: SPINAsmLanguageServer, params: lsp.DidChangeTextDocumentParams
):
    """Run diagnostics on changed document."""
    await ls.get_parser(params.text_document.uri)


@server.feature(lsp.TEXT_DOCUMENT_DID_SAVE)
async def did_save(ls: SPINAsmLanguageServer, params: lsp.DidSaveTextDocumentParams):
    """Run diagnostics on saved document."""
    await ls.get_parser(params.text_document.uri)


@server.feature(lsp.TEXT_DOCUMENT_DID_OPEN)
async def did_open(ls: SPINAsmLanguageServer, params: lsp.DidOpenTextDocumentParams):
    """Run diagnostics on open document."""
    await ls.get_parser(params.text_document.uri)


@server.feature(lsp.TEXT_DOCUMENT_DID_CLOSE)
def did_close(
    ls: SPINAsmLanguageServer, params: lsp.DidCloseTextDocumentParams
) -> None:
    """Clear the diagnostics on close."""
    ls.publish_diagnostics(params.text_document.uri, [])


def _get_defined_hover(stxt: str, parser: SPINAsmParser) -> str:
    """Get a hover message with the value of a defined variable or label."""
    # Check jmptbl first since labels are also defined in symtbl
    if stxt in parser.jmptbl:
        hover_definition = parser.jmptbl[stxt]
        return f"(label) {stxt}: Offset[{hover_definition}]"
    # Check constants next since they are also defined in symtbl
    if stxt in parser.constants:
        hover_definition = parser.symtbl[stxt]
        return f"(constant) {stxt}: Literal[{hover_definition}]"
    if stxt in parser.symtbl:
        hover_definition = parser.symtbl[stxt]
        return f"(variable) {stxt}: Literal[{hover_definition}]"

    return ""


@server.feature(lsp.TEXT_DOCUMENT_HOVER)
async def hover(ls: SPINAsmLanguageServer, params: lsp.HoverParams) -> lsp.Hover | None:
    """Retrieve documentation from symbols on hover."""
    parser = await ls.get_parser(params.text_document.uri)

    if (token := parser.token_registry.get_token_at_position(params.position)) is None:
        return None

    if token.symbol["type"] in ("LABEL", "TARGET"):
        hover_msg = _get_defined_hover(str(token), parser=parser)

        return (
            None
            if not hover_msg
            else lsp.Hover(
                # Java markdown formatting happens to give the best color-coding for
                # hover messages
                contents={"language": "java", "value": f"{hover_msg}"},
                range=token.range,
            )
        )

    if token.symbol["type"] in ("ASSEMBLER", "MNEMONIC"):
        hover_msg = ls.documentation.get_markdown(str(token))

        return (
            None
            if not hover_msg
            else lsp.Hover(
                contents=lsp.MarkupContent(
                    kind=lsp.MarkupKind.Markdown, value=hover_msg
                ),
                range=token.range,
            )
        )

    return None


@server.feature(lsp.TEXT_DOCUMENT_COMPLETION)
async def completions(
    ls: SPINAsmLanguageServer, params: lsp.CompletionParams
) -> lsp.CompletionList:
    """Returns completion items."""
    parser = await ls.get_parser(params.text_document.uri)

    symbol_completions = [
        lsp.CompletionItem(
            label=symbol,
            kind=lsp.CompletionItemKind.Constant
            if symbol in parser.constants
            else lsp.CompletionItemKind.Variable
            if symbol in parser.symtbl
            else lsp.CompletionItemKind.Module,
            detail=_get_defined_hover(symbol, parser=parser),
        )
        for symbol in {**parser.symtbl, **parser.jmptbl}
    ]

    opcode_completions = [
        lsp.CompletionItem(
            label=opcode,
            kind=lsp.CompletionItemKind.Function,
            detail="(opcode)",
            documentation=lsp.MarkupContent(
                kind=lsp.MarkupKind.Markdown,
                value=ls.documentation.get_markdown(opcode),
            ),
        )
        for opcode in [k.upper() for k in ls.documentation.instructions]
    ]

    assembler_completions = [
        lsp.CompletionItem(
            label=assembler,
            kind=lsp.CompletionItemKind.Operator,
            detail="(assembler)",
            documentation=lsp.MarkupContent(
                kind=lsp.MarkupKind.Markdown,
                value=ls.documentation.get_markdown(assembler),
            ),
        )
        for assembler in [k.upper() for k in ls.documentation.assemblers]
    ]

    return lsp.CompletionList(
        is_incomplete=False,
        items=symbol_completions + opcode_completions + assembler_completions,
    )


@server.feature(lsp.TEXT_DOCUMENT_DEFINITION)
async def definition(
    ls: SPINAsmLanguageServer, params: lsp.DefinitionParams
) -> lsp.Location | None:
    """Returns the definition location of a symbol."""
    parser = await ls.get_parser(params.text_document.uri)

    document = ls.workspace.get_text_document(params.text_document.uri)

    if (token := parser.token_registry.get_token_at_position(params.position)) is None:
        return None

    # Definitions should be checked against the base token name, ignoring address
    # modifiers.
    base_token = token.without_address_modifier()

    if str(base_token) not in parser.definitions:
        return None

    return lsp.Location(
        uri=document.uri,
        range=parser.definitions[str(base_token)],
    )


@server.feature(lsp.TEXT_DOCUMENT_DOCUMENT_SYMBOL)
async def document_symbol_definitions(
    ls: SPINAsmLanguageServer, params: lsp.DocumentSymbolParams
) -> list[lsp.DocumentSymbol]:
    """Returns the definitions of all symbols in the document."""
    parser = await ls.get_parser(params.text_document.uri)

    return [
        lsp.DocumentSymbol(
            name=symbol,
            kind=lsp.SymbolKind.Module
            if symbol in parser.jmptbl
            # There's no need to check for constants here since they aren't included
            # in the parser definitions.
            else lsp.SymbolKind.Variable,
            range=definition,
            selection_range=definition,
        )
        for symbol, definition in parser.definitions.items()
    ]


@server.feature(lsp.TEXT_DOCUMENT_PREPARE_RENAME)
async def prepare_rename(ls: SPINAsmLanguageServer, params: lsp.PrepareRenameParams):
    """Called by the client to determine if renaming the symbol at the given location
    is a valid operation."""
    parser = await ls.get_parser(params.text_document.uri)

    if (token := parser.token_registry.get_token_at_position(params.position)) is None:
        return None

    # Renaming is checked against the base token name, ignoring address modifiers.
    base_token = token.without_address_modifier()

    # Only user-defined labels should support renaming
    if str(base_token) not in parser.definitions:
        ls.info(f"Can't rename non-user defined token {base_token}.")
        return None

    return lsp.PrepareRenameResult_Type2(default_behavior=True)


@server.feature(
    lsp.TEXT_DOCUMENT_RENAME, options=lsp.RenameOptions(prepare_provider=True)
)
async def rename(
    ls: SPINAsmLanguageServer, params: lsp.RenameParams
) -> lsp.WorkspaceEdit:
    parser = await ls.get_parser(params.text_document.uri)

    if (token := parser.token_registry.get_token_at_position(params.position)) is None:
        return None

    # Ignore address modifiers so that e.g. we can rename `Delay` by renaming `Delay#`
    base_token = token.without_address_modifier()
    matching_tokens = parser.token_registry.get_matching_tokens(str(base_token))

    edits = [lsp.TextEdit(t.range, new_text=params.new_name) for t in matching_tokens]
    return lsp.WorkspaceEdit(changes={params.text_document.uri: edits})


@server.feature(lsp.TEXT_DOCUMENT_REFERENCES)
async def references(
    ls: SPINAsmLanguageServer, params: lsp.ReferenceParams
) -> list[lsp.Location]:
    parser = await ls.get_parser(params.text_document.uri)

    if (token := parser.token_registry.get_token_at_position(params.position)) is None:
        return []

    # Ignore address modifiers so that e.g. we can find all variations of addresses,
    # e.g. `Delay` and `Delay#`
    base_token = token.without_address_modifier()
    matching_tokens = parser.token_registry.get_matching_tokens(str(base_token))

    return [
        lsp.Location(uri=params.text_document.uri, range=t.range)
        for t in matching_tokens
    ]


@server.feature(
    lsp.TEXT_DOCUMENT_SIGNATURE_HELP,
    options=lsp.SignatureHelpOptions(
        trigger_characters=[" "], retrigger_characters=[","]
    ),
)
async def signature_help(
    ls: SPINAsmLanguageServer, params: lsp.SignatureHelpParams
) -> lsp.SignatureHelp | None:
    parser = await ls.get_parser(params.text_document.uri)

    try:
        line_tokens = parser.token_registry._tokens_by_line[params.position.line]
    # Exit on empty line
    except KeyError:
        return None

    # Find all opcodes on the line that could have triggered the signature help. Ignore
    # opcodes that appear after the cursor, to avoid showing signature help prematurely.
    opcodes = [
        t
        for t in line_tokens
        if t.symbol["type"] == "MNEMONIC"
        and t.range.start.character < params.position.character
    ]
    if not opcodes:
        return None

    # We should never have more than one opcode on a line, but just in case, grab the
    # last one entered before the cursor.
    triggered_opcode = opcodes[-1]
    opcode = ls.documentation.get_instruction(str(triggered_opcode))
    if opcode is None:
        return None

    remaining_tokens = line_tokens[line_tokens.index(triggered_opcode) + 1 :]

    argseps = [t for t in remaining_tokens if t.symbol["type"] == "ARGSEP"]
    # The first argument separator to the right of the cursor is the active arg index.
    # If there are no separators right of the cursor, we either haven't entered an arg
    # yet or we're on the last arg.
    arg_idx = 0 if not argseps else len(opcode.args) - 1
    for i, argsep in enumerate(argseps):
        if argsep.range.end.character > params.position.character:
            arg_idx = i
            break

    # Special case for multi-word instructions that treat the first argument as part of
    # the instruction name. In that case, we need to shift to the previous argument to
    # account for the extra arg separator.
    if str(triggered_opcode) in MULTI_WORD_INSTRUCTIONS and arg_idx != -1:
        arg_idx -= 1

    signature = [lsp.ParameterInformation(label=arg.markdown) for arg in opcode.args]

    return lsp.SignatureHelp(
        signatures=[
            lsp.SignatureInformation(
                label=f"{opcode.name} {opcode.args.markdown}",
                parameters=signature,
                documentation=lsp.MarkupContent(
                    kind=lsp.MarkupKind.Markdown,
                    value=opcode.markdown,
                ),
            )
        ],
        active_signature=0,
        active_parameter=arg_idx,
    )


def start() -> None:
    server.start_io()


if __name__ == "__main__":
    start()
